#===============================================================================
#
# Copyright (C) 2007 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

import sys
from optparse import OptionParser

from svndump import __version, copy_dump_file
from common import create_svn_date_str
from file import SvnDumpFile

__doc__ = """Various tools."""

#-------------------------------------------------------------------------------
# check

class SvnDumpCvs2SvnFix:
    """
    A class for cvs2svn created dump files.
    """

    def __init__( self ):
        """
        Initialize.
        """

        self.__history = {}

    def execute( self, inputfile, outputfile ):
        """
        Fix a cvs2svn created dump file.

        @type inputfile: string
        @param inputfile: Name of the cvs2svn created dump file.
        @type outputfile: string
        @param outputfile: Name of the fixed dump file.
        """

        indump = SvnDumpFile()
        indump.open( inputfile )
        has_rev = indump.read_next_rev()
        outdump = SvnDumpFile()
        outdump.create_like( outputfile, indump )
        rc = 0
        self.__history = {}

        while has_rev and rc == 0:
            outdump.add_rev( indump.get_rev_props() )
            for node in indump.get_nodes_iter():
                msglist = self.__fix_node( indump.get_rev_nr(), node )
                if msglist != None:
                    rc = 1
                    print "Error in r%d node %s:" % \
                            ( indump.get_rev_nr(), node.get_path() )
                    for msg in msglist:
                        print "  " + msg
                    break
                outdump.add_node( node )
            has_rev = indump.read_next_rev()

        indump.close()
        return rc

    def __fix_node( self, revnr, node ):
        """
        Checks the action of a node and keeps it's history.

        @type revnr: int
        @param revnr: Current revision number.
        @type node: SvnDumpNode
        @param node: Current node.
        """
        path = node.get_path()
        action = node.get_action()
        kind = None
        if action == "add":
            # path must not exist
            if self.__node_kind( revnr, path ) != None:
                return [ "Node already exists." ]
            # parent must be a dir
            slash = path.rfind( "/" )
            if slash > 0:
                ppath = path[:slash]
                pkind = self.__node_kind( revnr, ppath )
                if pkind == None:
                    return [ "Parent doesn't exist." ]
                elif pkind != "dir":
                    return [ "Parent is not a directory." ]
            # copy-from must exist
            if node.has_copy_from():
                kind = self.__node_kind( node.get_copy_from_rev(),
                                         node.get_copy_from_path() )
                if kind == None:
                    frompath = "  r%d %s" % ( node.get_copy_from_rev(),
                            node.get_copy_from_path() )
                    return [ "Copy-from path doesn't exist.", frompath ]
            if node.get_kind() == "":
                # missing node kind, fix it!
                if kind == None:
                    return [ "Unable to fix node." ]
                node.set_kind( kind )
                print "Set kind '%s' in r%d for %s" % \
                        ( kind, revnr, node.get_path() )
            self.__add_node( revnr, node )
        elif action == "delete":
            # path must exist
            kind = self.__node_kind( revnr, path )
            if kind == None:
                return [ "Node doesn't exist." ]
            self.__delete_node( revnr, node )
        else:
            # path must exist
            kind = self.__node_kind( revnr, path )
            if kind == None:
                return [ "Node doesn't exist." ]
            # replace = delete & add; changes can be ignored
            if action == "replace" and node.has_copy_from():
                self.__delete_node( revnr, node )
                self.__add_node( revnr, node )
            if node.get_kind() == "":
                # missing node kind, fix it!
                if kind == None:
                    return [ "Unable to fix node." ]
                node.set_kind( kind )
                print "Set kind '%s' in r%d for %s" % \
                        ( kind, revnr, node.get_path() )
        return None

    def __node_kind( self, revnr, path ):
        """
        Returns the kind of a node if it exists, else None.

        @type revnr: int
        @param revnr: Current revision number.
        @type path: string
        @param path: Path of a node.
        @rtype: string
        @return: "dir" for dirs, "file" for files or None.
        """
        if not self.__history.has_key( path ):
            return None
        nodehist = self.__history[ path ]
        i = self.__rev_index( nodehist, revnr )
        if i == None:
            return None
        return nodehist[0]

    def __rev_index( self, nodehist, revnr ):
        """
        Returns the index into the node history or None.

        @type nodehist: list
        @param nodehist: History of a node.
        @type revnr: int
        @param revnr: Current revision number.
        """
        i = len(nodehist) - 1
        while i > 0 and revnr < nodehist[i][0]:
            i -= 1
        if i == 0:
            return None
        if revnr > nodehist[i][1] and nodehist[i][1] >= 0:
            return None
        return i

    def __add_node( self, revnr, node ):
        """
        Adds a node to the history, recursively if it has copy-from path/rev.

        @type revnr: int
        @param revnr: Current revision number.
        @type node: SvnDumpNode
        @param node: Node to add.
        """
        path = node.get_path()
        if not self.__history.has_key( path ):
            # create revision list for path
            self.__history[ path ] = [ ( node.get_kind() ) ]
        # add revision range
        self.__history[ path ].append( [ revnr, -1 ] )
        kind = self.__history[ path ][0]
        # continue only if it's a dir with copy-from
        if kind == "file" or not node.has_copy_from():
            return
        # recursive copy
        cfpath = node.get_copy_from_path() + "/"
        cfpathlen = len(cfpath)
        cfrev = node.get_copy_from_rev()
        path += "/"
        for cfnodepath in self.__history.keys()[:]:
            if cfnodepath.startswith( cfpath ):
                cfnodehist = self.__history[cfnodepath]
                i = self.__rev_index( cfnodehist, cfrev )
                if i != None:
                    npath = path + cfnodepath[cfpathlen:]
                    # add new path
                    if not self.__history.has_key( npath ):
                        # create revision list for npath
                        self.__history[ npath ] = [ cfnodehist[0] ]
                    # add revision range
                    self.__history[ npath ].append( [ revnr, -1 ] )


    def __delete_node( self, revnr, node ):
        """
        Deletes a node from the history, recursively if it is a directory.

        @type revnr: int
        @param revnr: Current revision number.
        @type node: SvnDumpNode
        @param node: Node to add.
        """
        # set end revision
        path = node.get_path()
        self.__history[ path ][-1][1] = revnr - 1
        kind = self.__history[ path ][0]
        # continue only if it's a dir
        if kind == "file":
            return
        # recursive delete
        path += "/"
        for nodepath in self.__history.keys()[:]:
            if nodepath.startswith( path ):
                nodehist = self.__history[nodepath]
                if nodehist[-1][1] == -1:
                    nodehist[-1][1] = revnr - 1

def svndump_cvs2svnfix_cmdline( appname, args ):
    """
    Parses the commandline and executes the cvs2svnfix.

    Usage:

        >>> svndump_cvs2svnfix_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: int
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] inputfile outputfile" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    c2sfix = SvnDumpCvs2SvnFix()
    (options, args) = parser.parse_args( args )

    if len(args) != 2:
        print "Please specify exactly one input and one output file name."
        return 1

    return c2sfix.execute( args[0], args[1] )




#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

import calendar
import time

try:
    import hashlib
except ImportError:
    import md5 as hashlib

__doc__ = """Common functions and classes."""

# perhaps some of these date functions can be replaced by
# functions provided by the svn python bindings +++++
# <sussman> it's our own string format, in libsvn_subr/time.c
# <sussman> svn_time_[to|from]_cstring()
def parse_svn_date_str( dateStr ):
    """
    Parse a svn date string and return a tuple containing time_t and micros.
    """

    if len(dateStr) != 27:
        return (0,0)
    if dateStr[19] != "." or dateStr[26] != "Z":
        return (0,0)
    dat = time.strptime( dateStr[:19], "%Y-%m-%dT%H:%M:%S" )
    return ( int(calendar.timegm(dat)), int( dateStr[20:26] ) )

def create_svn_date_str( dateTuple ):
    """
    Creates a svn date string from a tuple containing time_t and micros.
    """

    dat = time.gmtime( dateTuple[0] )
    dstr = time.strftime( "%Y-%m-%dT%H:%M:%S", dat )
    mstr = ".%06dZ" % ( dateTuple[1] )
    return dstr + mstr

def is_valid_md5_string( md5 ):
    """
    Checks a md5 string.

    @type md5: object
    @param md5: Index or key.
    @rtype: bool
    @return: True if the string looks like an md5 sum.
    """

    if len( md5 ) != 32:
        return False
    if md5.lower().strip( "0123456789abcdef" ) != "":
        return False
    return True

class SvnDumpException( Exception ):
    """A simple exception class."""
    
    def __init__( self, text ):
        self.text = text

    def __str__( self ):
        return self.text

class ListDictIter:
    """
    Iterator class used by ListDict.
    """

    def __init__( self, listdict, type ):
        self.__listdict = listdict
        self.__type = type
        self.__index = 0

    def __iter__( self ):
        return self

    def next( self ):
        index = self.__index
        if index >= len( self.__listdict ):
            raise StopIteration()
        self.__index = index+1
        if self.__type == 1:
            return self.__listdict.key( index )
        elif self.__type == 2:
            return self.__listdict[ index ]
        else:
            return self.__listdict.item( index )

class ListDict( dict ):
    """
    A mix of list and dict.

    If the key is an int this class acts like a list else like a dict.
    """

    def __init__( self ):
        """
        Initialize.
        """
        dict.__init__( self )
        self.__index = []

    def __delitem__( self, key ):
        """
        Removes the key/value pair for the specified key or index.

        @type key: object
        @param key: Index or key.
        """
        if type( key ) is int:
            index = key
            key = self.__index[index]
        else:
            index = self.__index.index( key )
        del self.__index[index]
        dict.__delitem__( self, key )

    def __getitem__( self, key ):
        """
        Returns the value for the specified key or index if it's an int.

        @type key: object
        @param key: Index or key.
        @rtype: object
        @return: An object.
        """
        if type( key ) is int:
            key = self.__index[key]
        return dict.__getitem__( self, key )

    def __iter__( self ):
        """
        Returns an iterator returning key/value tuples ordered by index.

        @rtype: iterator
        @return: An iterator over the items.
        """
        return ListDictIter( self, 1 )

    def __setitem__( self, key, value ):
        """
        Adds a key/value pair or replaces the value if the key already exists.

        The key may be an int (=index) when replacing a value.

        @type key: object
        @param key: Key or index if replacing a value.
        @type value: object
        @param value: A value.
        """
        if type( key ) is int:
            key = self.__index[key]
        if not self.has_key( key ):
            self.__index.append( key )
        dict.__setitem__( self, key, value )

    def clear( self ):
        """
        Clears this ListDict.
        """

        dict.clear( self )
        self.__index = []

    def item( self, index ):
        """
        Returns the key/value tuple for the given index.

        @type index: int
        @param index: Index of the item (tuple).
        @rtype: tuple
        @return: An item (key/value pair).
        """
        key = self.__index[ index ]
        return ( key, dict.__getitem__( self, key ) )

    def items( self ):
        """
        Returns a list of key/value tuples ordered by index.

        @rtype: list
        @return: A list of values.
        """
        ret = []
        for key in self.__index:
            ret.append( ( key, dict.__getitem__( self, key ) ) )
        return ret

    def iteritems( self ):
        """
        Returns an iterator returning key/value tuples ordered by index.

        @rtype: iterator
        @return: An iterator over the items.
        """
        return ListDictIter( self, 0 )

    def iterkeys( self ):
        """
        Returns an iterator returning keys ordered by index.

        @rtype: iterator
        @return: An iterator over the keys.
        """
        return ListDictIter( self, 1 )

    def itervalues( self ):
        """
        Returns an iterator returning values ordered by index.

        @rtype: iterator
        @return: An iterator over the values.
        """
        return ListDictIter( self, 2 )

    def key( self, index ):
        """
        Returns the key for the given index.

        @type index: int
        @param index: Index of the key.
        @rtype: object
        @return: A key.
        """
        return self.__index[ index ]

    def keys( self ):
        """
        Returns a list of keys ordered by index.

        @rtype: list
        @return: A list of keys.
        """
        return self.__index

    def values( self ):
        """
        Returns a list of values ordered by index.

        @rtype: list
        @return: A list of values.
        """
        ret = []
        for key in self.__index:
            ret.append( dict.__getitem__( self, key ) )
        return ret

def sdt_md5():
    """
    Returns a new md5 object.

    For compatibility with python <2.5.
    """
    return hashlib.md5()




#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
# Copyright (C) 2007 CommProve, Inc (Eli Carter <eli.carter@commprove.com>)
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

from optparse import OptionParser
import os
import re
import sys

from svndump import __version, copy_dump_file, SvnDumpFile

class RevisionPropertyTransformer:
    """
    A class for transforming the revision properties of a dump file class.
    """

    def __init__( self, propertyName, regexStr, replaceTemplate ):
        """
        Creates a RevisionPropertyTransformer class.

        @type propertyName: string
        @param propertyName: Name of the property to transform.
        @type regexStr: string
        @param regexStr: The regular expression to match the value against.
        @type replaceTemplate: string
        @param replaceTemplate: The replacement string (may contain group references, e.g. \1).
        """
        self.__property_name = propertyName
        self.__pattern = re.compile(regexStr)
        self.__replace_template = replaceTemplate
        
    def transform( self, dump ):
        if dump.has_rev_prop( self.__property_name ):
            value = dump.get_rev_prop_value( self.__property_name )
            matcher = self.__pattern.match(value)
            if matcher:
                replace_str = matcher.expand( self.__replace_template )
                dump.set_rev_prop_value( self.__property_name, replace_str )

def svndump_transform_revprop_cmdline( appname, args ):
    """
    Parses the commandline and executes the transformation.

    Usage:

        >>> svndump_transform_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s propname regex replace source destination" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    (options, args) = parser.parse_args( args )

    if len( args ) != 5:
        print "specify exactly one propname to transform, one regex to match the value against,\none replacement string, one source dump file and one destination dump file."
        return 1

    copy_dump_file( args[3], args[4],  RevisionPropertyTransformer( args[0], args[1], args[2] ) )
    return 0

class PropertyTransformer:
    """
    A class for transforming the properties of a dump file class.
    """

    def __init__( self, propertyName, regexStr, replaceTemplate ):
        """
        Creates a RevisionPropertyTransformer class.

        @type propertyName: string
        @param propertyName: Name of the property to transform.
        @type regexStr: string
        @param regexStr: The regular expression to match the value against.
        @type replaceTemplate: string
        @param replaceTemplate: The replacement string (may contain group references, e.g. \1).
        """
        self.__property_name = propertyName
        self.__pattern = re.compile(regexStr, re.M)
        self.__replace_template = replaceTemplate

    def transform( self, dump ):
        for node in dump.get_nodes_iter():
            value = node.get_property(self.__property_name)
            if value != None:
                newvalue = self.__pattern.sub(self.__replace_template, value)
                node.set_property( self.__property_name, newvalue )

def svndump_transform_prop_cmdline( appname, args ):
    """
    Parses the commandline and executes the transformation.

    Usage:

        >>> svndump_transform_prop_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s propname regex replace source destination" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    (options, args) = parser.parse_args( args )

    if len( args ) != 5:
        print "specify exactly one propname to transform, one regex to match the value against,\none replacement string, one source dump file and one destination dump file."
        return 1

    copy_dump_file( args[3], args[4],  PropertyTransformer( args[0], args[1], args[2] ) )
    return 0


class SvnConfigParser:
    """
    A class similar to ConfigParser which actually works with SVN's config.
    """

    def __init__( self, filename ):
        """
        Initialize the object and read the config file.

        @type filename: string
        @param filename: Name of the config file.
        """
        self._sections = {}
        ifd = open( filename, "r" )
        section = None
        key = None
        val = None
        for line in ifd:
            line = line.rstrip()
            sline = line.lstrip()
            if line != sline:
                # continuation, append to the value
                if key != None:
                    val += " " + sline
            elif key != None:
                # end of continuation, store the key/value pair
                if section != None:
                    self._sections[section][key] = val
                key = None
            if len(sline) == 0 or sline[0] in ( ';', '#' ):
                # empty line or comment line
                pass
            elif line[0] == '[' and line[-1] == ']':
                # new section
                section = line[1:-1]
                self._sections[section] = {}
            else:
                ceq = line.find( "=" )
                cco = line.find( ":" )
                if ceq < 0 or (cco < ceq and cco >= 0):
                    ceq = cco
                if ceq >= 0:
                    # key/value pair
                    key = line[0:ceq].rstrip()
                    val = line[ceq+1:].lstrip()
        # store the last key/value pair
        if section != None and key != None:
            self._sections[section][key] = val

    def get( self, section, key ):
        """
        Returns the value for the specified key in the specified section.

        @type section: string
        @param section: Name of the section.
        @type key: string
        @param key: Key in the section.
        @rtype: string
        @return: Value for the specified key or empty string if not found.
        """
        if not self._sections.has_key( section ):
            return ""
        sdict = self._sections[section]
        if not sdict.has_key( key ):
            return ""
        return sdict[key]

    def items( self, section ):
        """
        Returns all key/value pairs in the specified section.

        @type section: string
        @param section: Name of the section.
        @rtype: ((key,value)...)
        @return: List of key/value pairs.
        """
        if not self._sections.has_key( section ):
            return []
        return self._sections[section].items()


class ApplyAutoprops:
    """
    A class for applying auto-props to a Subversion dump file.
    """

    def __init__( self, inputfilename, outputfilename, configfile ):
        """
        Initializes the ApplyAutoprops object.

        @type inputfilename: string
        @param inputfilename: Name of the input dump file.
        @type outputfilename: string
        @param outputfilename: Name of the output dump file.
        @type configfile: string
        @param configfile: Name of the Subversion config file.
        """
        self.configfile = configfile
        self.inputfilename = inputfilename
        self.outputfilename = outputfilename
        self.autoprops = []

    def apply( self ):
        """
        Applies the auto-props.
        """
        try:
            self._read_config()
            inDump = SvnDumpFile()
            outDump = SvnDumpFile()
            inDump.open( self.inputfilename )
            inDump.read_next_rev()
            outDump.create_like( self.outputfilename, inDump )
            while inDump.has_revision():
                print "revision %d" % inDump.get_rev_nr();
                outDump.add_rev( inDump.get_rev_props() )
                for index in range( 0, inDump.get_node_count() ):
                    node = inDump.get_node( index )
                    action = node.get_action()
                    if action in ( "add", "replace" ):
                        self._set_properties( node )
                    elif action == "change" and node.has_properties():
                        self._set_properties( node )
                    outDump.add_node( node )
                inDump.read_next_rev()
            inDump.close()
            outDump.close()
        except Exception, ex:
            print "Error:", ex
            return 1
        return 0

    def _set_properties( self, node ):
        """
        Set the auto-props.

        @type node: SvnDumpNode
        @param node: Node to set the properties on.
        """

        name = node.get_name()
        propkeys = ""
        for regex, properties in self.autoprops:
            if regex.match( name ):
                for pname, pval in properties:
                    node.set_property( pname, pval )
                    propkeys += ", " + pname
        if len(propkeys) > 0:
            print "  " + node.get_path()
            print "    set " + propkeys[2:]

    def _read_config( self ):
        """
        Reads the auto-props config.
        """
        cfg = SvnConfigParser( self.configfile )
        for key, value in cfg.items( "auto-props" ):
            regex = self._make_regex( key )
            properties = self._split_properties( value )
            self.autoprops.append( ( regex, properties ) )

    def _make_regex( self, expr ):
        """
        Convert the apr_fnmatch expression into a regular expression.

        @type expr: string
        @param expr: An apr_fnmatch expression.
        @rtype: re.Regex
        @return: Compiled regular expression.
        """
        replacements = {
                "\\": "\\\\",
                "(":  "\\(",
                ")":  "\\)",
                "{":  "\\{",
                "}":  "\\}",
                "|":  "\\|",
                "^":  "\\^",
                "$":  "\\$",
                "+":  "\\+",
                ".":  "\\.",
                "?":  ".",
                "*":  ".*",
        }
        rexpr = ""
        for c in expr:
            if c in replacements:
                c = replacements[c]
            rexpr += c
        return re.compile( "^%s$" % rexpr )

    def _split_properties( self, propstring ):
        """
        Splits the value of a key in the auto-props section.

        @type propstring: string
        @param propstring: A property string.
        @rtype: ((str,str)...)
        @return: Tuple containing property name and value tuples.
        """
        properties = []
        for property in propstring.split( ";" ):
            namevalue = property.split( "=", 1 )
            if len(namevalue) == 1:
                namevalue.append( "" )
            name = namevalue[0].strip()
            value = namevalue[1].strip()
            if value == "" and name in ( "svn:executable", "svn:needs-lock" ):
                value = "*"
            properties.append( ( name, value ) )
        return tuple( properties )


class SVNConfigFile:
    '''
    Determine a default SVN configuration file for the given system type.
    '''
    def __init__( self, envvar=None, pathelements=None, platform=None ):
        if platform is None:
            platform = sys.platform

        if platform == 'win32':
            self.envsep = '%'
            self.envstr = lambda: '%s%s%s' % (self.envsep, self.envvar, self.envsep)
            if envvar is None:
                envvar = 'APPDATA'
            if pathelements is None:
                pathelements = ['Subversion', 'config']
        else:
            self.envsep = '$'
            self.envstr = lambda: '%s%s' % (self.envsep, self.envvar)
            if envvar is None:
                envvar = 'HOME'
            if pathelements is None:
                pathelements = ['.subversion', 'config']

        self.envvar = envvar
        self.pathelements = pathelements

    def path(self):
        '''
        System compatible representation of the path (with the environment
        variable expanded.)
        '''
        home = os.environ[self.envvar]
        return os.path.join(home, *self.pathelements)

    def __str__(self):
        '''
        Human readable string representation of the path.
        '''
        return os.path.join(self.envstr(), *self.pathelements)


def svndump_apply_autoprops_cmdline( appname, args ):
    """
    Parses the commandline and applies the automatic properties.

    Usage:

        >>> svndump_apply_autoprops_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    svnconf = SVNConfigFile()

    usage = "usage: %s [options] inputdump outputdump" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    parser.add_option( "--config-file",
                       action="store", dest="configfile", default=None,
                       help="Subversion config file (default: %s)." % (svnconf, ))

    (options, args) = parser.parse_args( args )

    if len(args) != 2:
        print "Please specify exactly one input and one output dump file."
        return 1

    configfile = options.configfile
    if configfile == None:
        configfile = svnconf.path()

    aa = ApplyAutoprops( args[0], args[1], configfile )
    return aa.apply()




#!/usr/bin/env python
#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
# Copyright (C) 2007 CommProve, Inc. (Eli Carter <eli.carter@commprove.com>)
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

import os
import sys
import tempfile
import random
import string

from svndump import __version
from optparse import OptionParser
from file import SvnDumpFile
from __init__ import copy_dump_file

class SanitizeDumpFile(object):
    def __init__(self, options):
        self.__options = options
        self.sanitize_salt = self.salthex_to_salt(options.salt)
        print "Using salt %s" % (options.salt, )
        self.sanitized_authors = []

    def transform(self, dump):
        """The dump object passed to this method has been set to the revision
        we're to transform"""
        rev_props = dump.get_rev_props()
        self.sanitize_rev_props(rev_props)
        for node in dump.get_nodes_iter():
            node = self.sanitize_node(node)

    def salthex_to_salt(self, salthex):
        bytehex = []
        for i in range(0,len(salthex),2):
            bytehex.append(salthex[i:i+2])
        bytes = [chr(string.atoi(h, 16)) for h in bytehex]
        return ''.join(bytes)

    def sanitize_rev_props(self, rev_props ):
        for k in rev_props.iterkeys():
            if self.__options.logs and k == 'svn:log':
                rev_props['svn:log'] = self.hash(rev_props['svn:log'])
            elif self.__options.usernames and k == 'svn:author':
                author = rev_props['svn:author']
                try:
                    i = self.sanitized_authors.index(author)
                except ValueError:
                    i = len(self.sanitized_authors)
                    self.sanitized_authors.append(author)
                rev_props['svn:author'] = "author%s" % (i, )
            elif k == 'svn:date':
                pass # leave this for now.
            else:
                print "Couldn't sanitize %s: \"%s\"" % (k, rev_props[k])
        return rev_props

    # Sanitize the data using a salted md5sum
    def hash(self, data ):
        return hashlib.md5(self.sanitize_salt+data).hexdigest()

    def sanitize_path(self, path ):
        parts = path.split('/')
        sparts = []
        for part in parts:
            if part in self.__options.filename_excludes:
                spart = part
            else:
                spart = self.hash(part)
            sparts.append(spart)
        spath = '/'.join(sparts)
        return spath

    def sanitize_node(self, node ):
        if self.__options.file_data_method != "none" and node.has_text():
            # Write the original data to a temp file
            (fd, origname) = tempfile.mkstemp(prefix="svndumptool")
            origdatafile = os.fdopen(fd, "wb+")
            node.write_text_to_file(origdatafile)
            origdatafile.seek(0)
            # New data tempfile
            (fd, newname) = tempfile.mkstemp(prefix="svndumptool")
            newdatafile = os.fdopen(fd, "wb+")
            if self.__options.file_data_method == "whole":
                # Calculate the salted md5sum
                md5sum = sdt_md5()
                md5sum.update(self.sanitize_salt)
                while True:
                    data = origdatafile.read(1024**2)
                    if not len(data):
                        break
                    md5sum.update(data)
                # Write the sanitized data to another temp file.
                newdatafile.write("%s\n" % md5sum.hexdigest())
            elif self.__options.file_data_method == "line":
                # Write the sanitized data to another temp file.
                newdatafile.writelines(["%s\n" % self.hash(line) for line in origdatafile.readlines()])
            else:
                assert False, "self.__options.file_data_method has impossible value \"%s\"" % (self.__options.file_data_method, )
            # clean up new tempfile handle
            newdatafile.close()
            # clean up the original tempfile
            origdatafile.close()
            os.remove(origname)
            # Set the new content.
            node.set_text_file(newname, delete=True)

        if self.__options.filenames and node.get_path():
            node.set_path(self.sanitize_path(node.get_path()))
        if self.__options.filenames and node.has_copy_from():
            path = node.get_copy_from_path()
            rev = node.get_copy_from_rev()
            path = self.sanitize_path(path)
            node.set_copy_from(path, rev)
        # TODO: Properties?
        return node

def generate_salthex():
    salt = ''.join(["%02x" % (random.randrange(0,256), ) for x in range(8)])
    return salt

def svndump_sanitize_cmdline( appname, args ):
    """
    Parses the commandline and executes the sanitization.

    Usage:

        >>> svndump_sanitize_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] source destination" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    parser.add_option("-f", "--no-file-data",
        help="Do not sanitize file data.  (Equivalent to --file-data=none.)",
        action="store_const", const="none", dest="file_data_method", default=True)
    parser.add_option("-m", "--file-data",
        help="Method to sanitize file data: whole, line, none.  Default is whole.",
        type="choice", choices=["whole", "line", "none"],
        action="store", dest="file_data_method", default="whole")
    parser.add_option("-n", "--no-filenames",
        help="Do not sanitize filenames",
        action="store_false", dest="filenames", default=True)
    parser.add_option("-e", "--exclude-filename",
        help="Do not sanitize this filename.  May be used multiple times.",
        action="append", dest="filename_excludes", default=[])
    parser.add_option("-u", "--no-usernames",
        help="Do not sanitize usernames",
        action="store_false", dest="usernames", default=True)
    parser.add_option("-l", "--no-logs",
        help="Do not sanitize log messages",
        action="store_false", dest="logs", default=True)
    random_salt = generate_salthex()
    parser.add_option("-s", "--salt",
        help="Specify the salt to use in hex",
        dest="salt", default=random_salt)

    (options, args) = parser.parse_args( args )

    if len( args ) != 2:
        print "specify exactly one source and one destination dump file."
        return 1

    sanitizer = SanitizeDumpFile(options)
    copy_dump_file( args[0], args[1], sanitizer )
    return 0





#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

import sys
from optparse import OptionParser

from svndump import __version, copy_dump_file
from common import create_svn_date_str
from file import SvnDumpFileWithHistory, SvnDumpFile

__doc__ = """Various tools."""

#-------------------------------------------------------------------------------
# copy

def svndump_copy_cmdline( appname, args ):
    """
    Parses the commandline and executes the copy.

    Usage:

        >>> svndump_copy_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] source destination" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    (options, args) = parser.parse_args( args )

    if len( args ) != 2:
        print "specify exactly one source and one destination dump file."
        return 1

    copy_dump_file( args[0], args[1] )
    return 0


#-------------------------------------------------------------------------------
# export

class SvnDumpExport:
    """
    A class for exporting files from a svn dump files.
    """

    def __init__( self ):
        """
        Initialize.
        """

        # files to export, { rev -> { repos-path -> filename, ... }, ... }
        self.__exports = {}

    def add_export_file( self, revnr, repospath, filename ):
        """
        Add a file to export.

        @type revnr: integer
        @param revnr: Revision number.
        @type repospath: string
        @param repospath: Path of the file in the repository.
        @type filename: string
        @param filename: Name of the exported file.
        """

        if repospath.startswith('/'):
            repospath = repospath[1:]
        if not self.__exports.has_key( revnr ):
            self.__exports[revnr] = {}
        self.__exports[revnr][repospath] = filename

    def has_exports( self ):
        '''
        Were files listed for export?

        @rtype: boolean
        @return: True means there are files listed to
                 export, False means there are not.
        '''
        return len(self.__exports)>0

    def execute( self, dumpfilename, directory ):
        """
        Executes the export.

        @type dumpfilename: string
        @param dumpfilename: Name of the svn dump file.
        @type directory: string
        @param directory: Directory to store the exported files.
        """

        dump = SvnDumpFile()
        dump.open( dumpfilename )

        while dump.read_next_rev():
            revnr = dump.get_rev_nr()
            if self.__exports.has_key( revnr ):
                for path, filename in self.__exports[revnr].iteritems():
                    print "r%-6d %s" % ( revnr, path )
                    nodes = dump.get_nodes_by_path( path, "ACR" )
                    saved = False
                    for node in nodes:
                        if node.has_text():
                            outfile = open( filename, "wb" )
                            node.write_text_to_file( outfile )
                            outfile.close()
                            saved = True
                            print "  saved as %s" % filename
                    if not saved:
                        if len(nodes) == 0:
                            print "  not found"
                        else:
                            print "  has no text"
        dump.close()
        return 0

def __svndump_export_opt_e( option, opt, value, parser, *args ):
    """
    Option parser callback for rename '-r from to'.
    """
    export = args[0]
    export.add_export_file( int( value[0] ), value[1], value[2] )

def svndump_export_cmdline( appname, args ):
    """
    Parses the commandline and executes the export.

    Usage:

        >>> svndump_export_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] dumpfile" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    export = SvnDumpExport()
    parser.add_option( "-e", "--export",
                       action="callback", callback=__svndump_export_opt_e,
                       callback_args=(export,),
                       dest=" rev repospath filename",
                       nargs=3, type="string",
                       help="adds a file to export." )
    parser.add_option( "-d", "--directory",
                       action="store", type="string",
                       dest="dir", default=".",
                       help="set the directory for the exported files." )
    (options, args) = parser.parse_args( args )

    if not export.has_exports():
        print >>sys.stderr, "Warning: there are no files specified to export."
        print >>sys.stderr, "         Use -e/--export to specify files"
        print >>sys.stderr
        parser.print_help(file=sys.stderr)
        return 1
    return export.execute( args[0], options.dir )


#-------------------------------------------------------------------------------
# check

class SvnDumpCheck:
    """
    A class for checking svn dump files.
    """

    def __init__( self ):
        """
        Initialize.
        """

        # check actions
        self.__check_actions = False
        # check dates
        self.__check_dates = False
        # check md5 sums
        self.__check_md5 = False
        # verbose output
        self.__verbose = False

    def set_check_actions( self, docheck ):
        """
        Set the check actions flag to the given value.

        @type docheck: bool
        @param docheck: New value for the flag.
        """

        self.__check_actions = docheck

    def set_check_dates( self, docheck ):
        """
        Set the check dates flag to the given value.

        @type docheck: bool
        @param docheck: New value for the flag.
        """

        self.__check_dates = docheck

    def set_check_md5( self, docheck ):
        """
        Set the check md5 sums flag to the given value.

        @type docheck: bool
        @param docheck: New value for the flag.
        """

        self.__check_md5 = docheck

    def set_verbose( self, doverbose ):
        """
        Set for verbose output mode
        """

        self.__verbose = doverbose

    def execute( self, dumpfilename ):
        """
        Check a dump file.

        @type dumpfilename: string
        @param dumpfilename: Name of the file to check.
        """

        print "\nChecking file " + dumpfilename
        dump = SvnDumpFileWithHistory()
        if self.__check_actions:
            dump.set_check_actions(True)
        if self.__check_dates:
            dump.set_check_dates(True)
        if self.__check_md5:
            dump.set_check_md5(True)
        dump.open( dumpfilename )
        rc = 0

        while dump.read_next_rev():
            self.__next_rev()
            revnr = dump.get_rev_nr()
            if self.__verbose:
                self.__print_rev( revnr )
            if self.__print_rev_errors( dump ):
                rc = 1
            for node in dump.get_nodes_iter():
                self.__next_node()
                if self.__verbose:
                    self.__print_node( revnr, node )
                    self.__print_action( node )
                if self.__print_node_errors( dump, node ):
                    rc = 1
        dump.close()
        print [ "OK", "Not OK" ][ rc ]
        return rc

    def __next_rev( self ):
        """
        Clears the rev_printed flag.
        """
        self.__rev_printed = False

    def __print_rev( self, revnr ):
        """
        Prints the revision number if not already done.

        @type revnr: int
        @param revnr: Current revision number.
        """
        if not self.__rev_printed:
            self.__rev_printed = True
            print "  Revision %d" % revnr

    def __next_node( self ):
        """
        Clears the node_printed flag.
        """
        self.__node_printed = False
        self.__action_printed = False

    def __print_node( self, revnr, node ):
        """
        Prints the revision number and node path if not already done.

        @type revnr: int
        @param revnr: Current revision number.
        @type node: SvnDumpNode
        @param node: Current node.
        """
        if not self.__node_printed:
            self.__node_printed = True
            self.__print_rev( revnr )
            print "    Node: %s" % node.get_path()

    def __print_action( self, node ):
        """
        Prints the node action if not already done.

        @type node: SvnDumpNode
        @param node: Current node.
        """
        if not self.__action_printed:
            self.__action_printed = True
            actionmsg = "    Action: %s" % node.get_action()
            if node.has_copy_from():
                actionmsg += ", copied from r%d %s" % (
                        node.get_copy_from_rev(),
                        node.get_copy_from_path() )
            print actionmsg

    def __print_rev_errors( self, dump ):
        """
        Prints all revision errors for the current dump revision

        @type dump: SvnDumpFile
        @param dump: Current dump file.
        """

        rc = 0
        errlist = dump.get_rev_errors()
        if errlist is None:
            return rc
        revnr = dump.get_rev_nr()
        for err in errlist:
            if err[0] == SvnDumpFile.ERR_REV_DATE_OLDER:
                rc = 1
                self.print_rev( revnr )
                revdate = parse_svn_date_str( err[1][0] )
                prevdate = parse_svn_date_str( err[1][1] )
                print "    rev date: %s  %10d.%06d" % (
                    err[1][0], revdate[0], revdate[1] )
                print "    previous: %s  %10d.%06d" % (
                    err[1][1], prevdate[0], prevdate[1] )
        return rc

    def __print_node_errors( self, dump, node ):
        """
        Prints all node errors for the current dump node

        @type dump: SvnDumpFile
        @param dump: Current dump file.
        """

        rc = 0
        errlist = dump.get_rev_errors()
        if errlist is None:
            return rc
        revnr = dump.get_rev_nr()
        for err in errlist:
            if node.get_path() == err[1][0]:
                rc = 1
                self.__print_node( revnr, node )
                if err[0] == SvnDumpFile.ERR_NODE_MD5_FAIL:
                    print "      ERROR - md5 calc: %s" % err[1][0]
                    print "        diff than md5 node: %s" % err[1][1]
                if err[0] == SvnDumpFile.ERR_NODE_EXISTS:
                    print "      ERROR - Node already exists."
                if err[0] == SvnDumpFile.ERR_NODE_NO_PARENT:
                    print "      ERROR - Parent doesn't exist."
                if err[0] == SvnDumpFile.ERR_NODE_PARENT_NOT_DIR:
                    print "      ERROR - Parent is not a directory."
                if err[0] == SvnDumpFile.ERR_NODE_NO_COPY_SRC:
                    print "      ERROR - Copy-from path doesn't exist." \
                            "  r%d %s" % ( err[1][2], err[1][3] )
                if err[0] == SvnDumpFile.ERR_NODE_GONE:
                    print "      ERROR - Node doesn't exist."
        return rc

def svndump_check_cmdline( appname, args ):
    """
    Parses the commandline and executes the check.

    Usage:

        >>> svndump_check_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] dumpfiles..." % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    check = SvnDumpCheck()
    parser.add_option( "-a", "--check-actions",
                       action="store_true", dest="check_actions", default=False,
                       help="check actions like add/change/delete" )
    parser.add_option( "-d", "--check-dates",
                       action="store_true", dest="check_dates", default=False,
                       help="check that svn:date increases" )
    parser.add_option( "-m", "--check-md5",
                       action="store_true", dest="check_md5", default=False,
                       help="check md5 sums of the files" )
    parser.add_option( "-A", "--all-checks",
                       action="store_true", dest="check_all", default=False,
                       help="do all checks" )
    parser.add_option( "-v", "--verbose",
                       action="store_true", dest="verbose", default=False,
                       help="verbose output" )
    (options, args) = parser.parse_args( args )

    checks = False
    if options.check_actions or options.check_all:
        check.set_check_actions( True )
        checks = True
    if options.check_dates or options.check_all:
        check.set_check_dates( True )
        checks = True
    if options.check_md5 or options.check_all:
        check.set_check_md5( True )
        checks = True
    if options.verbose:
        check.set_verbose( True )

    if not checks:
        print "Please specify at least one check option."
        return 0

    rc = 0
    for filename in args:
        if check.execute( filename ) != 0:
            rc = 1
    return rc

#-------------------------------------------------------------------------------
# ls

class SvnDumpLs:
    """
    A class for listing files in a dump.
    """

    def __init__( self, revNr ):
        """
        Initialize.
        """

        self.revNr = revNr
        if revNr == -1:
            self.revNr = 2000000000

    def execute( self, dumpfilename ):
        """
        Print file list of a dump file.

        @type dumpfilename: string
        @param dumpfilename: Name of the file to log.
        """

        # pass 1: search copy-from revisions
        dump = SvnDumpFile()
        dump.open( dumpfilename )
        copyfromrevs = {}
        filedict = {}

        while dump.read_next_rev():
            revnr = dump.get_rev_nr()
            if revnr > self.revNr:
                break
            for node in dump.get_nodes_iter():
                #action = actions[node.get_action()]
                action = node.get_action()
                path = node.get_path()
                if path == "" or path[0] != "/":
                    path = "/" + path
                if action == "add" and node.has_copy_from():
                    copyfromrevs[node.get_copy_from_rev()] = True
        dump.close()

        # pass 2: do the work
        dump = SvnDumpFile()
        dump.open( dumpfilename )
        filedict = {}

        prevrevnr = 0
        while dump.read_next_rev():
            revnr = dump.get_rev_nr()
            # loop over missing revisions
            prevrevnr += 1
            while prevrevnr < revnr:
                if copyfromrevs.has_key(prevrevnr):
                    copyfromrevs[prevrevnr] = filedict.keys()[:]
                prevrevnr += 1
            if revnr > self.revNr:
                break
            for node in dump.get_nodes_iter():
                action = node.get_action()
                path = node.get_path()
                if path == "" or path[0] != "/":
                    path = "/" + path
                if action == "add":
                    filedict[path] = path
                    if node.has_copy_from():
                        frompath = node.get_copy_from_path() + "/"
                        if frompath[0] != "/":
                            frompath = "/" + frompath
                        fromlen = len(frompath)
                        topath = path + "/"
                        for path in copyfromrevs[node.get_copy_from_rev()]:
                            if path.startswith( frompath ):
                                newpath = topath + path[fromlen:]
                                filedict[newpath] = newpath
                elif action == "delete":
                    del filedict[path]
                    if path[-1] != "/":
                        path = path + "/"
                    for subpath in filedict.keys()[:]:
                        if subpath.startswith( path ):
                            del filedict[subpath]
            if copyfromrevs.has_key(revnr):
                copyfromrevs[revnr] = filedict.keys()[:]
        dump.close()

        filelist = []
        for path in filedict:
            filelist.append( path )
        filelist.sort()
        for path in filelist:
            print path

        return 0

    def old_execute( self, dumpfilename ):
        """
        Print file list of a dump file.

        @type dumpfilename: string
        @param dumpfilename: Name of the file to log.
        """

        dump = SvnDumpFile()
        dump.open( dumpfilename )
        actions = { "add":"A", "change":"M", "delete":"D", "replace":"R" }
        lines = ""

        while dump.read_next_rev():
            revnr = dump.get_rev_nr()
            if revnr == self.revNr or self.revNr == -1:
                lines = ""
                for node in dump.get_nodes_iter():
                    action = actions[node.get_action()]
                    path = node.get_path()
                    if path == "" or path[0] != "/":
                        path = "/" + path
                    if node.has_copy_from():
                        fpath = node.get_copy_from_path()
                        frev = node.get_copy_from_rev()
                        if fpath == "" or fpath[0] != "/":
                            fpath = "/" + fpath
                        path += " (from %s:%d)" % ( fpath, frev )
                    lines += "   %s %s\n" % ( action, path )
                if revnr == self.revNr:
                    print lines,
                    lines = ""
                    break
        if len(lines) > 0:
            print lines,

        dump.close()
        return 0

def svndump_ls_cmdline( appname, args ):
    """
    Parses the commandline and executes the ls.

    Usage:

        >>> svndump_ls_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] dumpfiles..." % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    parser.add_option( "-r", "--revision",
                       action="store", type="int",
                       dest="revnr", default=-1,
                       help="revision number" )
    (options, args) = parser.parse_args( args )

    log = SvnDumpLs( options.revnr )

    if len(args) == 1:
        return log.execute( args[0] )
    elif len(args) == 0:
        print "Specify a dump file."
        return 1
    else:
        print "Specify only one dump file."
        return 1


#-------------------------------------------------------------------------------
# log

class SvnDumpLog:
    """
    A class for checking svn dump files.
    """

    def __init__( self ):
        """
        Initialize.
        """

        # verbose
        self.__verbose = False
        # revision range
        self.__from_rev = -1
        self.__to_rev = 2000000000

    def set_verbose( self, verbose ):
        """
        Set the verbose flag to the given value.

        @type verbose: bool
        @param verbose: New value for the flag.
        """

        self.__verbose = verbose

    def set_revision( self, revision ):
        """
        Set the revision range to the given value.

        @type revision: string
        @param revision: A revision number or a range (X:Y).
        """

        if revision == None:
            return True
        if len(revision) == 0:
            return True
        parts = revision.split( ":" )
        n = len(parts)
        if n == 1:
            try:
                self.__from_rev = int(parts[0])
            except ValueError:
                print "Wrong format of revision argument '%s'" % revision
                return False
            self.__to_rev = self.__from_rev
        elif n == 2:
            try:
                self.__from_rev = int(parts[0])
                if parts[1] != "HEAD":
                    self.__to_rev = int(parts[1])
            except ValueError:
                print "Wrong format of revision argument '%s'" % revision
                return False
        else:
            print "Wrong format of revision argument '%s'" % revision
            return False
        return True

    def execute( self, dumpfilename ):
        """
        Print log of a dump file.

        @type dumpfilename: string
        @param dumpfilename: Name of the file to log.
        """

        print "\n\n" + "=" * 72
        line = "-" * 72
        print "Dumpfile: " + dumpfilename
        dump = SvnDumpFile()
        dump.open( dumpfilename )
        actions = { "add":"A", "change":"M", "delete":"D", "replace":"R" }

        while dump.read_next_rev():
            revnr = dump.get_rev_nr()
            if revnr >= self.__from_rev and revnr <= self.__to_rev:
                author = dump.get_rev_author()
                date = dump.get_rev_date_str()
                log = dump.get_rev_log()
                linecnt = len( log.split( "\n" ) )
                lines = "%d line" % linecnt
                if linecnt > 1:
                    lines += "s"
                print line
                print "r%d | %s | %s | %s" % ( revnr, author, date, lines )
                if self.__verbose:
                    print "Changed paths:"
                    for node in dump.get_nodes_iter():
                        action = actions[node.get_action()]
                        path = node.get_path()
                        if path == "" or path[0] != "/":
                            path = "/" + path
                        if node.has_copy_from():
                            fpath = node.get_copy_from_path()
                            frev = node.get_copy_from_rev()
                            if fpath == "" or fpath[0] != "/":
                                fpath = "/" + fpath
                            path += " (from %s:%d)" % ( fpath, frev )
                        print "   %s %s" % ( action, path )
                print "\n" + log.rstrip() + "\n"

        print line
        dump.close()
        return 0

def svndump_log_cmdline( appname, args ):
    """
    Parses the commandline and executes the log.

    Usage:

        >>> svndump_log_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] dumpfiles..." % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    log = SvnDumpLog()
    parser.add_option( "-r", "--revision",
                       action="store", type="string",
                       dest="revision", default=None,
                       help="revision number or range (X:Y)" )
    parser.add_option( "-v", "--verbose",
                       action="store_true", dest="verbose", default=False,
                       help="verbose output" )
    (options, args) = parser.parse_args( args )

    log.set_verbose( options.verbose )
    if not log.set_revision( options.revision ):
        return 1

    rc = 0
    for filename in args:
        if log.execute( filename ) != 0:
            rc = 1
    return rc


#-------------------------------------------------------------------------------
# join

def join_dumpfiles( inputlist, outfilename ):
    """
    Joins dump files.

    @type inputlist: list
    @param inputlist: A list containing the input filenames.
    @type outfilename: string
    @param outfilename: Name of the output dump file.
    @rtype: int
    @return: 0 for success.
    """

    outdump = None
    noutrev = 0
    lastrev = -1
    for filename in inputlist:
        print "reading %s ..." % filename
        ninrev = 0
        indump = SvnDumpFile()
        indump.open( filename )
        hasrev = indump.read_next_rev()
        if hasrev:
            if outdump == None:
                outdump = SvnDumpFile()
                if indump.get_rev_nr() == 0:
                    # create new dump with revision 0
                    outdump.create_with_rev_0( outfilename, indump.get_uuid(),
                                indump.get_rev_date_str() )
                    hasrev = indump.read_next_rev()
                else:
                    # create new dump starting with the
                    # same revNr as the original dump
                    outdump.create_with_rev_n( outfilename, indump.get_uuid(),
                                indump.get_rev_nr() )
            else:
                # check rev number
                if indump.get_rev_nr() == 0:
                    hasrev = indump.read_next_rev()
                if hasrev:
                    if (lastrev + 1) != indump.get_rev_nr():
                        print "renumbering of revisions not supported."
                        print "last rev was %d, next is %d." % ( lastrev,
                            indump.get_rev_nr() )
                        indump.close()
                        outdump.close()
                        return 1
            while hasrev:
                outdump.add_rev_from_dump( indump )
                ninrev += 1
                lastrev = indump.get_rev_nr()
                hasrev = indump.read_next_rev()
        indump.close()
        print "  copied %d revisions." % ninrev
        noutrev += ninrev
    outdump.close()
    print "wrote %d revisions, last was r%d." % ( noutrev, lastrev )

def svndump_join_cmdline( appname, args ):
    """
    Parses the commandline and executes the join.

    Usage:

        >>> svndump_join_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s -o outputfile dumpfiles..." % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    parser.add_option( "-o", "--output-file",
                       action="store", type="string",
                       dest="outfile", default=None,
                       help="set the name of the output dump file." )
    (options, args) = parser.parse_args( args )

    if options.outfile == None:
        print "please specify the output dump file (option -o)."
        return 1
    if len(args) == 0:
        print "please specify at least one input dump file."
        return 1

    return join_dumpfiles( args, options.outfile )

#-------------------------------------------------------------------------------
# split

def split_dumpfiles( inputfilename, outlist ):
    """
    Splits a dump file.

    @type inputfilename: string
    @param inputfilename: Name of the input file.
    @type outlist: list
    @param outlist: List of tuples containing start revnr, end revnr and filename.
    @rtype: int
    @return: 0 for success.
    """

    if len(outlist) == 0:
        return 0

    outlist = outlist[:]
    outlist.sort()
    parallel = False
    for i in range( 0, len(outlist)-1 ):
        if outlist[i][1] > outlist[i+1][0]:
            parallel = True
            break

    if not parallel:
        indump = SvnDumpFile()
        indump.open( inputfilename )
        index = 0
        startrev = outlist[index][0]
        endrev = outlist[index][1]
        outfile = outlist[index][2]
        outdump = None
        while indump.read_next_rev():
            revnr = indump.get_rev_nr()
            if outdump == None:
                if revnr >= startrev:
                    outdump = SvnDumpFile()
                    if revnr == 0:
                        # create new dump with revision 0
                        outdump.create_with_rev_0( outfile, indump.get_uuid(),
                                    indump.get_rev_date_str() )
                    else:
                        # create new dump starting with the
                        # same revNr as the original dump
                        outdump.create_with_rev_n( outfile, indump.get_uuid(),
                                    indump.get_rev_nr() )
            if outdump != None:
                # have an output file, copy the revision if revnr > 0
                if revnr > 0:
                    outdump.add_rev_from_dump( indump )
                if revnr >= endrev:
                    # end revision reached
                    outdump.close()
                    outdump = None
                    index += 1
                    if index >= len(outlist):
                        # done.
                        break
                    # next range
                    startrev = outlist[index][0]
                    endrev = outlist[index][1]
                    outfile = outlist[index][2]
        if outdump != None:
            outdump.close()
        indump.close()
    else:
        print "overlapping revision ranges not supported (yet)."
        return 1
    return 0

def svndump_split_cmdline( appname, args ):
    """
    Parses the commandline and executes the split.

    Usage:

        >>> svndump_split_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s inputfile [startrev endrev filename]..." % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    (options, args) = parser.parse_args( args )

    if len(args) == 0:
        return 0
    if (len(args) % 3) != 1:
        print "illegal number of args."
        return 0

    infile = args[0]
    outlist = []
    for i in range( 1, len(args), 3 ):
        outlist.append( ( int(args[i]), int(args[i+1]), args[i+2] ) )

    return split_dumpfiles( infile, outlist )




#!/usr/local/bin/python2.7
#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

import sys

from svndump import __version
from svndump.cvs2svnfix import svndump_cvs2svnfix_cmdline
from svndump.diff import svndump_diff_cmdline
from svndump.edit import svndump_edit_cmdline
from svndump.eolfix import svndump_eol_fix_cmdline
from svndump.merge import svndump_merge_cmdline
from svndump.props import svndump_transform_revprop_cmdline, \
                          svndump_transform_prop_cmdline, \
                          svndump_apply_autoprops_cmdline
from svndump.sanitize import svndump_sanitize_cmdline
from svndump.tools import svndump_copy_cmdline, svndump_export_cmdline, \
                          svndump_check_cmdline, svndump_log_cmdline, \
                          svndump_ls_cmdline, \
                          svndump_join_cmdline, svndump_split_cmdline

__commands = {
    "apply-autoprops":      svndump_apply_autoprops_cmdline,
    "check":                svndump_check_cmdline,
    "copy":                 svndump_copy_cmdline,
    "cvs2svnfix":           svndump_cvs2svnfix_cmdline,
    "diff":                 svndump_diff_cmdline,
    "edit":                 svndump_edit_cmdline,
    "eolfix":               svndump_eol_fix_cmdline,
    "export":               svndump_export_cmdline,
    "join":                 svndump_join_cmdline,
    "log":                  svndump_log_cmdline,
    "ls":                   svndump_ls_cmdline,
    "merge":                svndump_merge_cmdline,
    "sanitize":             svndump_sanitize_cmdline,
    "split":                svndump_split_cmdline,
    "transform-prop":       svndump_transform_prop_cmdline,
    "transform-revprop":    svndump_transform_revprop_cmdline,
}

def __help( appname, args ):
    rc = 0
    if len(args) == 1 and __commands.has_key( args[0] ):
        __commands[args[0]]( appname + " " + args[0], [ "-h" ] )
    else:
        print ""
        print "svndumptool.py command [options]"
        print ""
        print "  commands:"
        print "    apply-autoprops      apply auto-props to added files"
        print "    check                check a dump file"
        print "    copy                 copy a dump file"
        print "    cvs2svnfix           fix a cvs2svn created dump file"
        print "    diff                 show differences between two dump files"
        print "    edit                 edit files in a dump file"
        print "    eolfix               fix EOL of text files in a dump"
        print "    export               export files from a dump file"
        print "    join                 join dump files"
        print "    log                  show the log of a dump file"
        print "    ls                   list files of a given revision"
        print "    merge                merge dump files"
        print "    sanitize             sanitize dump files"
        print "    split                split dump files"
        print "    transform-revprop    transform a revision property"
        print "    transform-prop       transform a node property"
        print "    --version            print the version"
        print ""
        print "  use 'svndumptool.py command -h' for help about the commands."
        print ""
    return rc

def __print_version( appname, args ):
    print appname + " " + __version
    return 0

if __name__ == '__main__':
    appname = sys.argv[0].replace( "\\", "/" )
    n = appname.rfind( "/" )
    if n >= 0:
        appname = appname[n+1:]
    pfx = appname[0:7]
    cmd = appname[7:-3]
    sfx = appname[-3:]
    func = __help;
    args = []
    argidx = 0
    if pfx == "svndump" and sfx == ".py" and __commands.has_key( cmd ):
        func = __commands[cmd]
        argidx = 1
    elif len( sys.argv ) > 1:
        cmd = sys.argv[1]
        if __commands.has_key( cmd ):
            func = __commands[cmd]
            appname += " " + cmd
        elif cmd == "--version":
            func = __print_version
        argidx = 2
    if argidx < len( sys.argv ):
        args = sys.argv[argidx:]
    sys.exit( func( appname, args ) )




#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

from optparse import OptionParser
import re

from svndump import __version
from file import SvnDumpFile
from node import SvnDumpNode
from common import sdt_md5

__doc__ = """Classes and functions for fixing EOL's in a dump file."""

def eolfix_callback_prop( dumpfile, node, textfiles ):
    """
    Callback function which checks for svn:eol-style property.

    Returns True if a conversion of the text is needed.

    B{WARNING: This function is not tested!}
    """

    # do we allready know that it is a textfile ?
    if textfiles.has_key( node.get_path() ):
        # +++ should i check mime-type here ?
        if node.get_action() == "delete":
            del textfiles[node.get_path()]
        return True

    # handle copy-from-path ??? +++

    # check properties
    properties = node.get_properties()
    if properties == None:
        return False
    if not properties.has_key("svn:eol-style"):
        return False

    # is a text file, add to the list
    textfiles[node.get_path()] = dumpfile.get_rev_nr()
    return True

def eolfix_callback_regexp( dumpfile, node, expressions ):
    """
    Callback function which checks the node path using a regexp list.
    
    Returns True if a conversion of the text is needed.
    """

    for searchpath, re in expressions:
        if searchpath:
            name = node.get_path()
        else:
            name = node.get_name()
        if re.search( name ) != None:
            return True
    return False

class SvnDumpEolFix:
    """
    A class for fixing mixed EOL style files in a svn dump file.
    """

    def __init__( self ):
        """
        Initialize.
        """

        # input file name
        self.__in_file = ""
        # output file name
        self.__out_file = ""
        # dry-run, do not create an output file
        self.__dry_run = True
        # (do not) set eol-style on text files
        self.__eol_style = None
        # fix types
        self.__fix = 0
        # fix types for rev/path pairs
        self.__fix_rev_path = {}
        # warning file obj
        self.__warning_file = None
        # count of warnings logged
        self.__warning_count = 0
        # temp directory
        self.__temp_dir = "./"

        # temp files
        self.__temp_file_max_nr = 0
        self.__temp_file_nr = 0

    def set_input_file( self, filename ):
        """
        Sets the input dump file name.

        @type filename: string
        @param filename: Name of the input file.
        """
        self.__in_file = filename

    def set_output_file( self, filename ):
        """
        Sets the output dump file name and clears the dry-run flag.

        @type filename: string
        @param filename: Name of the output file.
        """
        self.__out_file = filename
        self.__dry_run = False

    def set_mode_prop( self ):
        """
        Sets mode to 'prop'.

        In this mode SvnDumpEolFix assumes text files have the property
        svn:eol-style set.

        B{See also:} eolfix_callback_prop
        """
        self.__is_text_file = eolfix_callback_prop
        self.__is_text_file_params = {}

    def set_mode_regexp( self, expressions ):
        """
        Sets mode to 'regexp'.

        In this mode every file which matches at least one of the regexps
        is treated as text file.

        B{See also:} eolfix_callback_regexp

        @type expressions: list( string )
        @param expressions: A list of regular expressions.
        """
        self.__is_text_file = eolfix_callback_regexp
        self.__is_text_file_params = []
        for expr in expressions:
            searchpath = expr.find( "/" ) >= 0
            regex = re.compile( expr )
            self.__is_text_file_params.append( ( searchpath, regex ) )

    def set_mode_callback( self, callback, parameter ):
        """
        Sets mode to 'callback'.

        @type callback: function( SvnDumpFile, SvnDumpNode, parameter )
        @param callback: Callback function to check if conversion is needed.
        @type parameter: any
        @param parameter: A parameter given to the callback function.
        """

        self.__is_text_file = callback
        self.__is_text_file_params = parameter

    def set_eol_style( self, eolstyle ):
        """
        Enable/disable setting eol-style on text files.
        
        If eolstyle is None do not set svn:eol-style, else set it
        to the given value.

        @type eolstyle: string
        @param eolstyle: Value for the svn:eol-style property or None.
        """
        self.__eol_style = eolstyle

    def set_fix_options( self, fix ):
        """
        Set what to fix.

        @type fix: string
        @param fix: A string containing comma separated options.
            Valid options are:
             - 'CRLF':  replace CRLF by LF
             - 'CR':    replace CR by LF
             - 'RemCR': remove CR
        """
        self.__fix = self.__parse_fix_option( fix )

    def set_fix_for_rev_file( self, fixrevfile ):
        """
        Set what to fix for a given revision/file.

        @type fixrevfile: string
        @param fixrevfile: A string containing colon separated the
            fix option, revsision number and path of a file.
        """
        parts = fixrevfile.split( ":", 2 )
        if len( parts ) != 3:
            print "wrong number of fiels for fixrevfile option."
            return
        key = ( int( parts[1] ), parts[2] )
        self.__fix_rev_path[key] = self.__parse_fix_option( parts[0] )
        print key, self.__fix_rev_path[key]

    def __parse_fix_option( self, fixstr ):
        """
        Parses a string containing comma separated fix options.

        @type fixstr: string
        @param fixstr: Fix option string.
        @rtype: integer
        @return: Fix options.
        """
        fix = 0
        for f in fixstr.split( ',' ):
            if f == "CRLF":
                fix |= 1
            elif f == "CR":
                fix |= 2
            elif f == "RemCR":
                fix |= 4
        return fix

    def set_warning_file( self, warnfile ):
        """
        Sets the filename for writing warnings into.

        @type warnfile: string
        @param warnfile: Name of the warnings file.
        """
        if self.__warning_file != None:
            self.__warning_file.close()
        self.__warning_count = 0
        self.__warning_file = open( warnfile, "w" )
        self.__warning_file.write(
                "#/bin/sh\n" +
                "\n" +
                "# the following statements fetch the files for which\n" +
                "# warnings have been issued from the repository\n" +
                "\n" +
                "# name of the svn executable\n" +
                "SVN=svn\n" +
                "# repository URL\n" +
                "REPOS=file:///tmp/repos\n" +
                "\n" )

    def set_temp_dir( self, tmpdir ):
        """
        Sets the the directory for temporary files.

        @type tmpdir: string
        @param tmpdir: Name of the tmp dir.
        """
        if tmpdir[-1] != "/":
            tmpdir += "/"
        self.__temp_dir = tmpdir

    def execute( self ):
        """
        Executes the EolFix.
        """

        # +++ catch exception and return errorcode
        srcdmp = SvnDumpFile()
        srcdmp.open( self.__in_file )

        dstdmp = None

        hasrev = srcdmp.read_next_rev()
        if hasrev:
            if not self.__dry_run:
                dstdmp = SvnDumpFile()
                if srcdmp.get_rev_nr() == 0:
                    # create new dump with revision 0
                    dstdmp.create_with_rev_0( self.__out_file,
                                              srcdmp.get_uuid(),
                                              srcdmp.get_rev_date_str() )
                    hasrev = srcdmp.read_next_rev()
                else:
                    # create new dump starting with the same revNr
                    # as the input dump file
                    dstdmp.create_with_rev_n( self.__out_file,
                                              srcdmp.get_uuid(),
                                              srcdmp.get_rev_nr() )
            # now copy all the revisions
            while hasrev:
                print "\n\n*** r%d ***\n" % srcdmp.get_rev_nr()
                self.__process_rev( srcdmp, dstdmp )
                hasrev = srcdmp.read_next_rev()
        if self.__warning_file != None:
            self.__warning_file.write(
                    "\n\n# %d warnings\n" % self.__warning_count )
            self.__warning_file.close()
            self.__warning_file = None
            self.__warning_count = 0

    def __process_rev( self, srcdmp, dstdmp ):
        """
        Process one revision.

        @type srcdmp: SvnDumpFile
        @param srcdmp: The source dump file.
        @type dstdmp: SvnDumpFile
        @param dstdmp: The destination dump file.
        """

        # clear temp file nr (overwrite old files)
        self.__temp_file_nr = 0

        # add revision and revprops
        if dstdmp != None:
            dstdmp.add_rev( srcdmp.get_rev_props() )

        # process nodes
        index = 0
        nodeCount = srcdmp.get_node_count()
        while index < nodeCount:
            node = srcdmp.get_node( index )
            print "  '%s'" % node.get_path()
            istextfile = False
            if node.get_kind() == 'dir':
                print "    directory, ignored"
            else:
                istextfile = self.__is_text_file( srcdmp, node,
                                                  self.__is_text_file_params )
                if istextfile:
                    node = self.__convert_eol( node, srcdmp.get_rev_nr() )
                else:
                    print "    unselected file, ignored"
            # +++ is node.has_properties a good enough test?
            # maybe i have to save the properties of each node and
            # use them here?
            if self.__eol_style != None and node.has_properties():
                if istextfile:
                    node.set_property( "svn:eol-style", self.__eol_style )
            if dstdmp != None:
                dstdmp.add_node( node )
            index = index + 1

    def __convert_eol( self, node, revnr ):
        """
        Convert EOL of a node.

        @type node: SvnDumpNode
        @param node: The node to convert.
        @type revnr: integer
        @param revnr: The current revision number.
        @rtype: SvnDumpNode
        @return: The converted node.
        """

        if not node.has_text():
            # no text
            print "    selected file, no text changes"
            return node

        # search for CR
        need_conv = False
        handle = node.text_open()
        data = node.text_read( handle )
        while len(data) > 0:
            if data.find( "\r" ) != -1:
                # found one, need to convert the file
                need_conv = True
                break
            data = node.text_read( handle )
        fix = self.__fix
        if need_conv:
            # special fix option for rev/file ?
            key = ( revnr, node.get_path() )
            if self.__fix_rev_path.has_key( key ):
                fix = self.__fix_rev_path[key]
            print "    selected file, convert (fix option %d)" % fix
            if self.__dry_run or fix == 0:
                # do not convert with --dry-run or when there's nothing to fix
                need_conv = False
        else:
            print "    selected file, no conversion required"
        if need_conv:
            # do the conversion
            node.text_reopen( handle )
            outfilename = self.__temp_file_name()
            outfile = open( outfilename, "wb" )
            outlen = 0
            md = sdt_md5()
            data = node.text_read( handle )
            carry = ""
            warning_printed = False
            while len(data) > 0:
                if len(carry) != 0:
                    data = carry + data
                n = len( data ) - 1
                carry = data[n]
                if carry == "\r":
                    data = data[:n]
                else:
                    carry = ""
                if fix & 1 != 0:
                    data = data.replace( "\r\n", "\n" )
                if fix & 2 != 0:
                    data = data.replace( "\r", "\n" )
                if fix & 4 != 0:
                    data = data.replace( "\r", "" )
                if not warning_printed and data.find( "\r" ) >= 0:
                    warning_printed = True
                    print "    WARNING: file still contains CR"
                    print "      file: '%s'" % node.get_path()
                    if self.__warning_file != None:
                        self.__warning_file.write(
                                "# WARNING: file still contains CR\n" )
                        file = node.get_path()
                        while file[0] == "/":
                            file = file[1:]
                        tmpfile = file.replace( "/", "__" )
                        cmd = '$SVN cat -r %d "$REPOS/%s" > "%s"\n' % \
                            ( revnr, node.get_path(), tmpfile )
                        self.__warning_file.write( cmd )
                        self.__warning_count += 1
                md.update( data )
                outfile.write( data )
                outlen = outlen + len( data )
                data = node.text_read( handle )
            if len( carry ) != 0:
                if fix & 2 != 0:
                    carry = "\n"
                elif fix & 4 != 0:
                    carry = ""
                outfile.write( carry )
                md.update( carry )
                outlen += len( carry )
            outfile.close()
            newnode = SvnDumpNode( node.get_path(), node.get_action(),
                                   node.get_kind() )
            if node.has_copy_from():
                newnode.set_copy_from( node.get_copy_from_path(),
                                       node.get_copy_from_rev() )
            if node.has_properties():
                newnode.set_properties( node.get_properties() )
            newnode.set_text_file( outfilename, outlen, md.hexdigest() )
        else:
            newnode = node

        node.text_close( handle )
        return newnode

    def __temp_file_name( self ):
        """
        Create temp file name.

        @rtype: string
        @return: A temp file name.
        """
        self.__temp_file_nr = self.__temp_file_nr + 1
        if self.__temp_file_nr > self.__temp_file_max_nr:
            self.__temp_file_max_nr = self.__temp_file_nr
        return "%stmpnode%d" % ( self.__temp_dir, self.__temp_file_nr )


def svndump_eol_fix_cmdline( appname, args ):
    """
    Parses the commandline and executes the eolfix.

    Usage:

        >>> svndump_eol_fix_cmdline( sys.argv[0], sys.argv[1:] )

    @type appname: string
    @param appname: Name of the application (used in help text).
    @type args: list( string )
    @param args: Commandline arguments.
    @rtype: integer
    @return: Return code (0 = OK).
    """

    usage = "usage: %s [options] src [dst]" % appname
    parser = OptionParser( usage=usage, version="%prog "+__version )
    parser.add_option( "-E", "--eol-style",
                       action="store", dest="eolstyle", default=None,
                       type="choice", choices=[ "native", "LF", "CRLF", "CR" ],
                       help="add svn:eol-style property to text files, "
                            "the value can be 'native', 'LF', 'CRLF' or 'CR'" )
    parser.add_option( "-f", "--fix",
                       action="store", dest="fix",
                       type="string", default="CRLF",
                       help="a comma separated list of what (and how) to fix, "
                            "can be a combination of "
                            "'CRLF', 'CR' and 'RemCR'. If 'CR' and 'RemCR' "
                            "both specified 'RemCR' is ignored. 'CRLF' and "
                            "'CR' mean replace them by LF, 'RemCR' means "
                            "remove CR's." )
    parser.add_option( "-F", "--fix-rev-path",
                       action="append", dest="fixrevpath",
                       type="string",
                       help="a colon separated list of fix option, revision "
                            "number and path of a file." )
    parser.add_option( "-r", "--regexp",
                       action="append", dest="regexp",
                       help="regexp for matching text file names or the " +
                            "full path if the expression contains a slash. " +
                            "This option can be specified more than once." )
    parser.add_option( "-t", "--temp-dir",
                       action="store", dest="tmpdir",
                       type="string",
                       help="directory for temporary files " +
                            "(default: './')." )
    parser.add_option( "-w", "--warn-file",
                       action="store", dest="warnfile",
                       type="string",
                       help="file for storing the warnings." )
    parser.add_option( "--dry-run",
                       action="store_true", dest="dry_run", default=False,
                       help="just show what would be done but don't do it" )
    (options, args) = parser.parse_args( args )

    eolfix = SvnDumpEolFix()

    if len( args ) < 1 or len( args ) > 2:
        print "please specify one source and optionally one destination file."
        return 1
    eolfix.set_input_file( args[0] )
    if len( args ) == 2:
        eolfix.set_output_file( args[1] )
    if options.regexp != None and len( options.regexp ) > 0:
        eolfix.set_mode_regexp( options.regexp )
    else:
        eolfix.set_mode_prop()
    if options.eolstyle != None:
        eolfix.set_eol_style( options.eolstyle )
    eolfix.set_fix_options( options.fix )
    if options.fixrevpath != None:
        for f in options.fixrevpath:
            eolfix.set_fix_for_rev_file( f )
    if options.tmpdir != None:
        eolfix.set_temp_dir( options.tmpdir )
    if options.warnfile != None:
        eolfix.set_warning_file( options.warnfile )

    eolfix.execute()
    return 0
        




#===============================================================================
#
# Copyright (C) 2003 Martin Furter <mf@rola.ch>
#
# This file is part of SvnDumpTool
#
# SvnDumpTool is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# SvnDumpTool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SvnDumpTool; see the file COPYING.  If not, write to
# the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
#===============================================================================

from os import stat, remove
from stat import ST_SIZE

from common import *

__doc__ = """SvnDumpNode class."""

class SvnDumpNode:
    """
    A node of a svn dump file.
    """

    def __init__( self, path, action, kind ):
        """
        Initializes a new SvnDumpNode.

        @type path: string
        @param path: Path of this node.
        @type action: string
        @param action: Action of this node.
            Valid actions are:
             - 'add'
             - 'delete'
             - 'change'
             - 'replace'
        @type kind: string
        @param kind: Kind of this node ('dir' or 'file').
            If action is 'delete' kind may be the empty string.
        """

        # check action
        if action != "add" and action != "delete" and \
           action != "change" and action != "replace":
            raise SvnDumpException, "Unknown action '%s'." % action
        # check kind
        if action != "change":
            if kind != "" and kind != "file" and kind != "dir":
                raise SvnDumpException, "Unknown kind '%s'" % kind
        else:
            if kind != "file" and kind != "dir":
                raise SvnDumpException, "Unknown kind '%s'" % kind
        # check path +++

        # path of this node relative to the repository root
        self.__path = path
        # action: 'add', 'change', 'delete' or 'replace'
        self.__action = action
        # kind: 'file', 'dir' or 'node' if not known
        self.__kind = kind
        # list of properties name=>value pairs
        self.__properties = None
        # length of the text (file data)
        self.__text_len = -1
        # md5 hash of the text
        self.__text_md5 = ""
        # the from path if copied else ""
        self.__copy_from_path = ""
        # the from revision if copied else 0
        self.__copy_from_rev = 0
        # offset into the file
        self.__file_offset = -1
        # name of the (temp) file
        self.__file_name = ""
        # delete the temp file
        self.__file_delete = False
        # the file object to read from
        self.__file_obj = None

    def __del__( self ):
        """
        Delete method, cleanup temp file if needed.
        """
        if self.__file_delete and self.__file_name != "":
            # delete temp file
            remove(self.__file_name)

    def get_path( self ):
        """
        Returns the path of this node.

        @rtype:  string
        @return: The path of this node.
        """
        return self.__path

    def set_path( self, path ):
        """
        Sets the path of this node.

        @type path: string
        @param path: New path of this node."""
        self.__path = path

    def get_name( self ):
        """
        Returns the name of this node.

        @rtype:  string
        @return: The name of this node.
        """
        return self.__path.split( "/" )[-1]

    def get_action( self ):
        """
        Returns the action of this node.

        @rtype:  string
        @return: Either 'add', 'change', 'delete' or 'replace'.
        """
        return self.__action

    def get_kind( self ):
        """
        Returns the kind of this node.

        @rtype: string
        @return: Either 'file', 'dir' or ''.
        """
        return self.__kind

    def get_property( self, name ):
        """
        Returns the  value of the property with the given name.

        If the property does not exist None is returned.

        @type name: string
        @param name: A property name.
        @rtype: string
        @return: Value of the property.
        """
        if self.__properties != None and self.__properties.has_key( name ):
            return self.__properties[name]
        else:
            return None

    def has_properties( self ):
        """
        Returns True if this node has properties.

        @rtype: bool
        @return: True if this node has properties.
        """
        return self.__properties != None

    def get_properties( self ):
        """
        Returns the properties as a dict.

        @rtype: dict( string -> string )
        @return: The properties of this node.
        """
        return self.__properties

    def has_text( self ):
        """
        Returns true when this node has text.

        @rtype: bool
        @return: True when this node has text.
        """
        return self.__text_len >= 0

    def get_text_length( self ):
        """
        Returns the length of the text.

        @rtype: integer
        @return: Length of the text.
        """
        return self.__text_len

    def has_md5( self ):
        """
        Returns true when this node has a MD5 sum.

        @rtype: bool
        @return: True when this node has a MD5 sum.
        """
        return len( self.__text_md5 ) > 0

    def get_text_md5( self ):
        """
        Returns the MD5 hash of the text.

        @rtype: string
        @return: MD5 sum of the text.
        """
        return self.__text_md5

    def has_copy_from( self ):
        """
        Returns True when this node has copy-from-path and copy-from-rev.

        @rtype: bool
        @return: True if this node has copy-from rev and path.
        """
        return self.__copy_from_rev > 0 and self.__copy_from_path != ""

    def get_copy_from_path( self ):
        """
        Returns the path the node has been copied from or an empty string.

        @rtype: string
        @return: copy-from-path.
        """
        return self.__copy_from_path

    def get_copy_from_rev( self ):
        """
        Returns the revision the node has been copied from or zero.

        @rtype: integer
        @return: copy-from-rev.
        """
        return self.__copy_from_rev

    def set_copy_from( self, path, revnr ):
        """
        Sets copy-from-path and copy-from-rev.

        Only nodes with action 'add' or 'replace' may have copy-from
        revision and path.

        @type path: string
        @param path: copy-from-path
        @type revnr: integer
        @param revnr: copy-from-rev
        """

        if self.__action != "add" and self.__action != "replace":
            raise SvnDumpException, "Cannot set copy-from for action '%s'" \
                    % self.__action
        self.__copy_from_path = path
        self.__copy_from_rev = revnr

    def set_kind( self, kind ):
        """
        Set the kind of this node.

        The kind can only be set if it was empty.

        @type kind: string
        @param kind: New kind, either 'file' or 'dir'.
        """
        if self.__kind != "":
            raise SvnDumpException, "Cannot change node kind"
        if kind != "file" and kind != "dir":
            raise SvnDumpException, "Unknown kind '%s'" % kind
        self.__kind = kind

    def set_property( self, name, value ):
        """
        Sets a property of this node.

        Nodes with action 'delete' cannot have properties.

        @type name: string
        @param name: Name of the property.
        @type value: string
        @param value: Value of the property.
        """

        if self.__action == "delete":
            raise SvnDumpException, "Cannot set properties for action '%s'" \
                    % self.__action
        if self.__properties == None:
            self.__properties = {}
        self.__properties[name] = value

    def del_property( self, name ):
        """
        Deletes a property of this node.

        @type name: string
        @param name: Name of the property to delete."""

        if self.__action == "delete":
            raise SvnDumpException, "Cannot delete properties for action '%s'" \
                    % self.__action
        if self.__properties != None:
            if self.__properties.has_key( name ):
                del self.__properties[name]
                if len( self.__properties ) == 0:
                    self.__properties = None

    def set_properties( self, properties ):
        """
        Sets the properties for this node.

        @type properties: dict( string -> string )
        @param properties: A dict containing the properties.
        """

        if self.__action == "delete":
            raise SvnDumpException, "Cannot set properties for action '%s'" \
                    % self.__action
        self.__properties = properties

    def set_text_file( self, filename, length=-1, md5="", delete=False ):
        """
        Sets the text for this node.

        The text will be read from the specified file.

        @type filename: string
        @param filename: Name of the file containing the text.
        @type length: integer, optional
        @param length: Length of the file.
        @type md5: string, optional
        @param md5: MD5 sum of the text if known.
        @type delete: bool
        @param delete: When True delete the file.
        """

        if self.__action == "delete":
            raise SvnDumpException, "Cannot set text for action '%s'" \
                    % self.__action
        if self.__kind != "file":
            raise SvnDumpException, "Cannot set text for kind '%s'" \
                    % self.__kind
        self.__file_name = filename
        self.__file_offset = 0
        # hmm, no destructors, how to delete that damn temp file ? +++
        self.__file_delete = delete
        if length == -1:
            length = stat( filename )[ST_SIZE]
        self.__text_len = length
        self.__text_md5 = md5
        if not is_valid_md5_string( md5 ):
            self.__calculate_md5()

    def set_text_fileobj( self, fileobj, offset, length, md5 ):
        """
        Sets the text for this node.

        The text will be read from the specified file object.

        @type fileobj: file object
        @param fileobj: A file object opened for reading and
            containing the text.
        @type offset: integer
        @param offset: Offset of the text.
        @type length: integer
        @param length: Length of the text.
        @type md5: string
        @param md5: MD5 sum of the text.
        """

        if self.__action == "delete":
            raise SvnDumpException, "Cannot set text for action '%s'" \
                    % self.__action
        if self.__kind != "file":
            raise SvnDumpException, "Cannot set text for kind '%s'" \
                    % self.__kind
        self.__file_obj = fileobj
        self.__file_offset = offset
        self.__text_len = length
        self.__text_md5 = md5
        #if !is_valid_md5_string( md5 ) or length == -1:
        #    self.__calculate_md5()

    def set_text_node( self, node ):
        """
        Sets the text for this node.

        The text will be that of the specified node.

        @type node: SvnDumpNode
        @param node: An other node.
        """

        if self.__action == "delete":
            raise SvnDumpException, "Cannot set text for action '%s'" \
                    % self.__action
        if self.__kind != "file":
            raise SvnDumpException, "Cannot set text for kind '%s'" \
                    % self.__kind
        self.__file_name = node.__file_name
        # dunno how to delete temp file so no special action here +++
        self.__file_delete = node.__file_delete
        self.__file_obj = node.__file_obj
        self.__file_offset = node.__file_offset
        self.__text_len = node.__text_len
        self.__text_md5 = node.__text_md5

    def write_text_to_file( self, outfile ):
        """
        Writes the text to the given file object.

        @type outfile: file object
        @param outfile: A file object opened for writing.
        """

        if self.__text_len == -1:
            raise SvnDumpException, "Node %s has no text" % self.__path
        if len(self.__file_name) > 0:
            self.__file_obj = open( self.__file_name, "rb" )
        else:
            self.__file_obj.seek( self.__file_offset )
        cnt = self.__text_len
        while cnt > 0:
            bcnt = cnt
            if bcnt > 16384:
                bcnt = 16384
            outfile.write( self.__file_obj.read( bcnt ) )
            cnt = cnt - bcnt
        if len(self.__file_name) > 0:
            self.__file_obj.close()
            self.__file_obj = None

    def text_open( self ):
        """
        Open text and return a handle for text read functions.

        Only one handle per dump file should be opened at the same time. A
        node for which the text has been set with set_text_node() virtually
        belongs also to the dump file of the node specified to
        set_text_node().

        Also while the handle is open write_text_to_file() should not be
        called.

        B{See also:} text_reopen(), text_read() and text_close().

        @rtype: 'opaque handle'
        @return: A handle for the text read functions.
        """

        if self.__text_len == -1:
            raise SvnDumpException, "node has no text"

        # create handle
        handle = {}
        if len(self.__file_name) > 0:
            handle["file_obj"] = open( self.__file_name, "rb" )
            handle["close"] = True
            handle["offset"] = 0
            handle["length"] = self.__text_len
            handle["pos"] = 0
        else:
            handle["file_obj"] = self.__file_obj
            handle["close"] = False
            handle["offset"] = self.__file_offset
            handle["length"] = self.__text_len
            handle["pos"] = 0
            self.__file_obj.seek( self.__file_offset )

        return handle

    def text_reopen( self, handle ):
        """
        Reopen the handle.

        Repositions the handle to the start of the text.

        B{See also:} text_open(), text_read() and text_close().

        @type handle: handle
        @param handle: A handle opened with text_open().
        """

        handle["file_obj"].seek( handle["offset"] )
        handle["pos"] = 0

    def text_read( self, handle, count=16384 ):
        """
        Read some text from a handle.

        B{See also:} text_open(), text_reopen() and text_close().

        @type handle: handle
        @param handle: A handle opened with text_open().
        @type count: integer, optional
        @param count: Count of bytes to read.
        @rtype: string
        @return: The data read.
        """

        # end of text ?
        if handle["pos"] >= handle["length"]:
            return ""

        # is more text requested than remains
        if (handle["pos"] + count) > handle["length"]:
            count = handle["length"] - handle["pos"]
        # read it
        data = handle["file_obj"].read( count )
        handle["pos"] = handle["pos"] + count
        return data

    def text_close( self, handle ):
        """
        Close the handle.

        B{See also:} text_open(), text_reopen() and text_read().

        @type handle: handle
        @param handle: A handle opened with text_open().
        """

        if handle["close"]:
            handle["file_obj"].close()

    def __calculate_md5( self ):
        """
        Calculates the md5 of the text of this node.
        """

        handle = self.text_open()
        md = sdt_md5()
        data = self.text_read( handle )
        n = 0
        while len(data) > 0:
            n = n + len(data)
            md.update( data )
            data = self.text_read( handle )
        self.__text_md5 = md.hexdigest()
        if self.__text_len == -1:
            self.__text_len = n
        self.text_close( handle )

